#include <assert.h>
#include <mbstring.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "avltree.h"
#include "common.h"

#undef abs

#define MSHOWERRHDL(expression) if (!(expression)) {\
        retres = 1;\
        printf("Error!\n");\
        show_and_clear_error();\
        goto lbl_finally;\
    }

static void destroy_tree_r(struct treenode *root);
static void print_tree_r(struct treenode *node, int level);
static bool insert_value_totree_1(struct treenode *rootroot, int val, struct treenode **ppnode);
static void update_tnheights_and_adjust_i(struct treenode *rootroot, struct treenode *newnode);
static void avl_ins_adjust(struct treenode *node, struct treenode *child, struct treenode *gchild, struct treenode *ggchild);
static void avl_ins_adjust_ll(struct treenode *node, struct treenode *child, bool flip);
static void avl_ins_adjust_lr(struct treenode *node, struct treenode *child, struct treenode *gchild, bool flip);
static struct treenode *swap_valueof_treenodetodel(struct treenode *node);
static void delete_node_fromtree_1(struct treenode *node);
static void update_tnheights_and_adjust_d(struct treenode *rootroot, struct treenode *node);
static void avl_del_adjust(struct treenode *node);
static void avl_del_adjust_2(struct treenode *node, struct treenode *child, bool flip);
static void avl_del_adjust_3(struct treenode *node, struct treenode *child, struct treenode *gchild, bool flip);
static int auto_get_tnheight(struct treenode *node);
static void swap_nodes(struct treenode *pnode1, struct treenode *pnode2);
static int calc_chlnbased_tnheight(struct treenode *node);

void init_treenode(struct treenode *node)
{
    node->parent = NULL;
    node->lchild = NULL;
    node->rchild = NULL;
    node->value = 0;
    node->height = 0;
}

void destroy_tree(struct treenode *rootroot)
{
    if (rootroot == NULL) {
        return;
    }
    destroy_tree_r(rootroot->lchild);
}

static void destroy_tree_r(struct treenode *root)
{
    if (root != NULL) {
        destroy_tree_r(root->lchild);
        destroy_tree_r(root->rchild);
        free(root);
    }
}

void print_tree(struct treenode *root)
{
    print_tree_r(root, 0);
}

static void print_tree_r(struct treenode *node, int level)
{
    int i;

    if (node != NULL) {
        print_tree_r(node->lchild, level + 1);
        for (i = 0; i < level; i++) {
            printf("\t");
        }
        printf("%d\n", node->value);
        if (node->lchild != NULL) {
            assert(node->lchild->parent == node);
        }
        if (node->rchild != NULL) {
            assert(node->rchild->parent == node);
        }
        assert(node->height == calc_chlnbased_tnheight(node));
        assert(abs(auto_get_tnheight(node->lchild) - auto_get_tnheight(
            node->rchild)) <= 1);
        print_tree_r(node->rchild, level + 1);
    } else {
        for (i = 0; i < level; i++) {
            printf("\t");
        }
        printf("$\n");
    }
}

bool insert_value_totree(struct treenode *rootroot, int val)
{
    bool retres = true;
    struct treenode *pnode;

    ERRHDL(insert_value_totree_1(rootroot, val, &pnode));
    update_tnheights_and_adjust_i(rootroot, pnode);
lbl_finally:
    return retres;
}

static bool insert_value_totree_1(struct treenode *rootroot, int val, struct treenode **ppnode)
{
    bool retres = true;
    struct treenode *pnode;
    struct treenode *pnode2;
    int lr; /* 0 - left, 1 - right */

    *ppnode = NULL;
    for (pnode = rootroot; pnode != NULL; ) {
        if (pnode == rootroot) {
            lr = 0;
        } else if (val <= pnode->value) {
            lr = 0;
        } else {
            lr = 1;
        }
        pnode2 = pnode;
        pnode = (lr == 0) ? pnode->lchild : pnode->rchild;
    }
    pnode = (struct treenode *)malloc(sizeof(struct treenode));
    RAISEERR(pnode != NULL, "insert_value_totree_1: malloc failed.\n");
    init_treenode(pnode);
    if (lr == 0) {
        pnode2->lchild = pnode;
    } else {
        pnode2->rchild = pnode;
    }
    pnode->parent = pnode2;
    pnode->value = val;
    *ppnode = pnode;
lbl_finally:
    return retres;
}

/**
 *  update_tnheights_and_adjust_i:
 *      Update tree node heights and adjust the tree after a node is inserted
 *      (what the last `i' stands for).
 */
static void update_tnheights_and_adjust_i(struct treenode *rootroot, struct treenode *newnode)
{
    struct treenode *pnode;
    struct treenode *pnode2;
    struct treenode *pnode3;
    struct treenode *pnode4;
    int ahdiff; /* absolute height difference */

    pnode2 = NULL;
    pnode3 = NULL;
    pnode4 = NULL;
    for (pnode = newnode; pnode != rootroot; pnode = pnode->parent) {
        /* NOTE: newnode->height will be calculated as 1. */
        pnode->height = calc_chlnbased_tnheight(pnode);
        ahdiff = abs(auto_get_tnheight(pnode->lchild) - auto_get_tnheight(
            pnode->rchild));
        if (ahdiff == 2) {
            avl_ins_adjust(pnode, pnode2, pnode3, pnode4);
            break;
        }
        assert(ahdiff <= 1);
        pnode4 = pnode3;
        pnode3 = pnode2;
        pnode2 = pnode;
    }
}

static void avl_ins_adjust(struct treenode *node, struct treenode *child, struct treenode *gchild, struct treenode *ggchild)
{
    /* NOTE: `child' and `gchild' must't be NULL */
    assert(child != NULL && gchild != NULL);
    if (child == node->lchild) {
        if (gchild == child->lchild) {
            /* LL */
            avl_ins_adjust_ll(node, child, false);
        } else if (gchild == child->rchild) {
            if (ggchild == NULL || ggchild == gchild->lchild || ggchild ==
                gchild->rchild)
            {
                /* LR, LRL or LRR: the same solution */
                avl_ins_adjust_lr(node, child, gchild, false);
            } else {
                assert(false);
            }
        }
    } else {
        assert(child == node->rchild);
        /* flipped */
        if (gchild == child->rchild) {
            avl_ins_adjust_ll(node, child, true);
        } else if (gchild == child->lchild) {
            if (ggchild == NULL || ggchild == gchild->rchild || ggchild ==
                gchild->lchild)
            {
                avl_ins_adjust_lr(node, child, gchild, true);
            } else {
                assert(false);
            }
        }
    }
}

static void avl_ins_adjust_ll(struct treenode *node, struct treenode *child, bool flip)
{
    if (!flip) {
        if (node->parent->lchild == node) {
            node->parent->lchild = child;
        } else {
            node->parent->rchild = child;
        }
        node->lchild = child->rchild;
        child->rchild = node;
        child->parent = node->parent;
        node->parent = child;
        if (node->lchild != NULL) {
            node->lchild->parent = node;
        }
    } else {
        if (node->parent->rchild == node) {
            node->parent->rchild = child;
        } else {
            node->parent->lchild = child;
        }
        node->rchild = child->lchild;
        child->lchild = node;
        child->parent = node->parent;
        node->parent = child;
        if (node->rchild != NULL) {
            node->rchild->parent = node;
        }
    }
    node->height = calc_chlnbased_tnheight(node);
    child->height = calc_chlnbased_tnheight(child);
    assert(abs(auto_get_tnheight(child->lchild) - auto_get_tnheight(
        child->rchild)) <= 1);
}

static void avl_ins_adjust_lr(struct treenode *node, struct treenode *child, struct treenode *gchild, bool flip)
{
    if (!flip) {
        if (node->parent->lchild == node) {
            node->parent->lchild = gchild;
        } else {
            node->parent->rchild = gchild;
        }
        gchild->parent = node->parent;
        node->parent = gchild;
        node->lchild = gchild->rchild;
        child->parent = gchild;
        child->rchild = gchild->lchild;
        gchild->lchild = child;
        gchild->rchild = node;
        if (node->lchild != NULL) {
            node->lchild->parent = node;
        }
        if (child->rchild != NULL) {
            child->rchild->parent = child;
        }
    } else {
        if (node->parent->rchild == node) {
            node->parent->rchild = gchild;
        } else {
            node->parent->lchild = gchild;
        }
        gchild->parent = node->parent;
        node->parent = gchild;
        node->rchild = gchild->lchild;
        child->parent = gchild;
        child->lchild = gchild->rchild;
        gchild->rchild = child;
        gchild->lchild = node;
        if (node->rchild != NULL) {
            node->rchild->parent = node;
        }
        if (child->lchild != NULL) {
            child->lchild->parent = child;
        }
    }
    child->height = calc_chlnbased_tnheight(child);
    node->height = calc_chlnbased_tnheight(node);
    gchild->height = calc_chlnbased_tnheight(gchild);
    assert(abs(auto_get_tnheight(child->lchild) - auto_get_tnheight(
        child->rchild)) <= 1);
    assert(abs(auto_get_tnheight(node->lchild) - auto_get_tnheight(
        node->rchild)) <= 1);
    assert(abs(auto_get_tnheight(gchild->lchild) - auto_get_tnheight(
        gchild->rchild)) <= 1);
}

void delete_value_fromtree(struct treenode *rootroot, int val)
{
    struct treenode *pnode;
    struct treenode *pnode2;

    pnode = find_value_intree(rootroot->lchild, val);
    if (pnode == NULL) {
        return;
    }
    pnode = swap_valueof_treenodetodel(pnode);
    pnode2 = pnode->parent;
    delete_node_fromtree_1(pnode);
    pnode = pnode2;
    update_tnheights_and_adjust_d(rootroot, pnode);
}

struct treenode *find_value_intree(struct treenode *root, int val)
{
    struct treenode *pnode;

    for (pnode = root; pnode != NULL; ) {
        if (pnode->value == val) {
            break;
        } else if (pnode->value > val) {
            pnode = pnode->lchild;
        } else {
            pnode = pnode->rchild;
        }
    }
    return pnode;
}

static struct treenode *swap_valueof_treenodetodel(struct treenode *node)
{
    struct treenode *pnode;

    if (node->lchild == NULL || node->rchild == NULL) {
        /* at the bottom; no need to swap */
        return node;
    }
    for (pnode = node->lchild; pnode->rchild != NULL; pnode = pnode->rchild) {
        ;
    }
    swap_nodes(node, pnode);
    return pnode;
}

static void delete_node_fromtree_1(struct treenode *node)
{
    struct treenode *pnode;
    struct treenode *nodechil; /* node's only child (left or right) or NULL */

    pnode = node->parent;
    assert(node->lchild == NULL || node->rchild == NULL);
    nodechil = (node->lchild != NULL) ? node->lchild : node->rchild;
    if (pnode->lchild == node) {
        pnode->lchild = nodechil;
    } else {
        pnode->rchild = nodechil;
    }
    if (nodechil != NULL) {
        nodechil->parent = pnode;
    }
    free(node);
}

/**
 *  update_tnheights_and_adjust_d:
 *      Update tree node heights and adjust the tree after a node is deleted
 *      (what the last `d' stands for).
 */
static void update_tnheights_and_adjust_d(struct treenode *rootroot, struct treenode *node)
{
    struct treenode *pnode;
    struct treenode *pnode2;
    int ahdiff; /* absolute height difference */

    for (pnode = node; pnode != rootroot; pnode = pnode2) {
        pnode->height = calc_chlnbased_tnheight(pnode);
        pnode2 = pnode->parent;
        ahdiff = abs(auto_get_tnheight(pnode->lchild) - auto_get_tnheight(
            pnode->rchild));
        if (ahdiff == 2) {
            avl_del_adjust(pnode);
        } else {
            assert(ahdiff <= 1);
        }
    }
}

static void avl_del_adjust(struct treenode *node)
{
    int l0diff; /* level 0 diff */
    int l1diff;

    l0diff = auto_get_tnheight(node->lchild) - auto_get_tnheight(node->rchild);
    if (l0diff == -2) {
        l1diff = auto_get_tnheight(node->rchild->lchild) - auto_get_tnheight(
            node->rchild->rchild);
        if (l1diff == 0 || l1diff == -1) {
            avl_del_adjust_2(node, node->rchild, false);
        } else {
            assert(l1diff == 1);
            avl_del_adjust_3(node, node->rchild, node->rchild->lchild, false);
        }
    } else {
        assert(l0diff == 2);
        /* flipped */
        l1diff = auto_get_tnheight(node->lchild->rchild) - auto_get_tnheight(
            node->lchild->lchild);
        if (l1diff == 0 || l1diff == -1) {
            avl_del_adjust_2(node, node->lchild, true);
        } else {
            assert(l1diff == 1);
            avl_del_adjust_3(node, node->lchild, node->lchild->rchild, true);
        }
    }
}

static void avl_del_adjust_2(struct treenode *node, struct treenode *child, bool flip)
{
    /* NOTE: avl_del_adjust_2 is so similar to avl_ins_adjust_ll so that we can
     * call it, just with a negative flip. */
    avl_ins_adjust_ll(node, child, (bool)!flip);
}

static void avl_del_adjust_3(struct treenode *node, struct treenode *child, struct treenode *gchild, bool flip)
{
    /* NOTE: like avl_del_adjust_2, avl_del_adjust_3 is also similar to
     * avl_ins_adjust_lr, just with a negative flip. */
    avl_ins_adjust_lr(node, child, gchild, (bool)!flip);
}

static int auto_get_tnheight(struct treenode *node)
{
    if (node == NULL) {
        return 0;
    } else {
        return node->height;
    }
}

static void swap_nodes(struct treenode *pnode1, struct treenode *pnode2)
{
    int val;

    val = pnode1->value;
    pnode1->value = pnode2->value;
    pnode2->value = val;
}

/* calc_chlnbased_tnheight: calculates children-based tree node height. */
static int calc_chlnbased_tnheight(struct treenode *node)
{
    return max_int(auto_get_tnheight(node->lchild) + 1, auto_get_tnheight(
        node->rchild) + 1);
}

