#include <assert.h>
#include <mbstring.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "avltree.h"
#include "avltree0.h"
#include "common.h"

#define MSHOWERRHDL(expression) if (!(expression)) {\
        retres = 1;\
        printf("Error!\n");\
        show_and_clear_error();\
        goto lbl_finally;\
    }

static bool do_things(void);
static bool accept_command(struct name *cmd, bool *eof);
static bool execute_command(struct treenode *rootroot, const char *cmd);
static bool parse_command(struct cmd_s *dest, char *src);

int main(void)
{
    int retres = 0; /* result to return */

    MSHOWERRHDL(do_things());
lbl_finally:
    return retres;
}

static bool do_things(void)
{
    bool retres = true;
    struct name cmd;
    /* rootroot: rootroot's lchild is the true root. */
    struct treenode *rootroot = NULL;
    bool eof;

    init_name(&cmd);
    rootroot = (struct treenode *)malloc(sizeof(struct treenode));
    RAISEERR(rootroot != NULL, "do_things: malloc failed.\n");
    init_treenode(rootroot);
    while (true) {
        ERRHDL(accept_command(&cmd, &eof));
        if (eof) {
            break;
        }
        ERRHDL(execute_command(rootroot, get_buffer_of_name(&cmd)));
    }
lbl_finally:
    destroy_name(&cmd);
    if (rootroot != NULL) {
        destroy_tree(rootroot);
        free(rootroot);
    }
    return retres;
}

static bool accept_command(struct name *cmd, bool *eof)
{
    enum { BUFSIZE = 1024 };
    bool retres = true;
    char *buffer = NULL;

    *eof = false;
    clear_name(cmd);
    buffer = (char *)malloc(BUFSIZE);
    RAISEERR(buffer != NULL, "accept_command: malloc failed.\n");
    if (fgets(buffer, BUFSIZE, stdin) == NULL) {
        *eof = true;
        goto lbl_finally;
    }
    remove_eol(buffer);
    ERRHDL(concat_name_str(cmd, buffer));
lbl_finally:
    if (buffer != NULL) {
        free(buffer);
    }
    return retres;
}

static bool execute_command(struct treenode *rootroot, const char *cmd)
{
    bool retres = true;
    char *dupcmd = NULL; /* duplicate */
    struct cmd_s parsed;

    dupcmd = (char *)malloc(get_min_power_of_two_ge(strlen(cmd) + 1));
    RAISEERR(dupcmd != NULL, "execute_command: malloc failed.\n");
    _mbscpy(dupcmd, cmd);
    ERRHDL(parse_command(&parsed, dupcmd));
    if (parsed.pcmd == NULL) {
        printf("execute_command: command not given.\n");
        goto lbl_finally;
    }
    if (_mbscmp(parsed.pcmd, "list") == 0) {
        print_tree(rootroot->lchild);
    } else if (_mbscmp(parsed.pcmd, "insert") == 0) {
        ERRHDL(insert_value_totree(rootroot, atoi(parsed.parg1)));
    } else if (_mbscmp(parsed.pcmd, "delete") == 0) {
        delete_value_fromtree(rootroot, atoi(parsed.parg1));
    } else if (_mbscmp(parsed.pcmd, "") == 0) { /* empty command */
        ;
    } else {
        printf("%s: not a command.\n", parsed.pcmd);
    }
lbl_finally:
    if (dupcmd != NULL) {
        free(dupcmd);
    }
    return retres;
}

static bool parse_command(struct cmd_s *dest, char *src)
{
    bool retres = true;
    char *p;

    dest->pcmd = NULL;
    dest->parg1 = NULL;
    for (p = src; *p != '\0' && isspace(*p); p = _mbsinc(p)) {
        ;
    }
    dest->pcmd = p;
    for (; *p != '\0' && !isspace(*p); p = _mbsinc(p)) {
        ;
    }
    *p = '\0';
    for (p++; *p != '\0' && isspace(*p); p = _mbsinc(p)) {
        ;
    }
    dest->parg1 = p;
    for (; *p != '\0' && !isspace(*p); p = _mbsinc(p)) {
        ;
    }
    *p = '\0';
lbl_finally:
    return retres;
}

void remove_eol(char *buffer)
{
    char *p;

    for (p = buffer; *p != '\0'; p = _mbsinc(p)) {
        if (*p == '\n') {
            *p = '\0';
            break;
        }
    }
}

