#if !defined(COMMON_H_20050906_INCLUDED_)
#define COMMON_H_20050906_INCLUDED_

#include <stdlib.h>
#include <windows.h>

/**
 *  common.h:
 *      Defines common interfaces for rosh.
 */

/**
 *  Multi-byte string usage guide:
 *      When using multi-byte strings, observe the following guidelines:
 *      1. When advancing or going back a character, use _mbsinc and _mbsdec.
 *      2. When representing beginning or end of a substring, use either a
 *         pointer to character or an index on bytes, rather than an index on
 *         multi-byte characters.
 */

#if !defined(true)
#define true 1
#endif
#if !defined(false)
#define false 0
#endif
typedef char bool;

#define ERRHDL(expression) if (!(expression)) {\
        retres = false;\
        goto lbl_finally;\
    }

#define SHOWERRHDL(expression) if (!(expression)) {\
        printf("Error!\n");\
        show_and_clear_error();\
        goto lbl_finally;\
    }

#define RAISEERR(expression, message) if (!(expression)) {\
        retres = false;\
        add_error_message(message);\
        goto lbl_finally;\
    }

#define QUENEXT(variable, quesize) (((variable) + 1) % (quesize))
#define QUEPREV(variable, quesize) (((variable) - 1 + (quesize)) % (quesize))

/* initialize { */

void initialize_on_run(void);

/* } initialize */

/* names { */

enum { NAME_PRIMBUFSIZE = 40 /* primary buffer size */ };

/**
 *  name:
 *      The structure represents not very long names. Too long names (longer
 *      than 1K characters) may hinder performance in some situations.
 *
 *  Members variables:
 *      prim_buf - the primary buffer, used directly to hold short names
 *      second_buf - a pointer to the secondary buffer. When it is not NULL, it
 *          is always used as the buffer, no longer how short the name is.
 *      length - the length of the name
 *      size - the size of the buffer
 *      destroyed - a boolean value inidicating whether the name structure is
 *          destroyed
 *
 *  NOTE:
 *      The string in the buffer should be null-terminated, and this should
 *      always be kept consistent with the `length' member.
 */
struct name
{
    char prim_buf[NAME_PRIMBUFSIZE];
    char *second_buf;
    int length;
    int size;
    bool destroyed;
};

void init_name(struct name *name);
bool concat_name(struct name *dest, struct name *src);
bool concat_name_str(struct name *name, const char *content);
char *get_buffer_of_name(struct name *name);
bool concat_name_part(struct name *dest, struct name *src, int begin, int end);
bool concat_name_str_part(struct name *dest, const char *begin, const char *end);
void trunc_name(struct name *dest, int new_len);
bool expand_name(struct name *name, int expected_len);
void clear_name(struct name *name);
void destroy_name(struct name *name);

/* } names */

/* calculate { */

int max_int(int a, int b);
int min_int(int a, int b);
bool queue_between(int testee, int begin, int end, int queue_size);
void re_place_last_in_order(void *base, size_t num, size_t width, int (__cdecl *compare)(const void *, const void *));
int get_min_power_of_two_ge(int n);

/* } calculate */

/* errors { */

enum { ERROR_MESSAGE_SIZE = 1024 };

extern char g_error_message[ERROR_MESSAGE_SIZE];

void add_error_message(const char *msg);
void clear_error_message(void);
void show_and_clear_error(void);

/* } errors */

#endif // !defined(COMMON_H_20050906_INCLUDED_)
